# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Boolean Ring
=============
"""
from typing import Dict

from residuated_binars.abelian_group import AbelianGroup
from residuated_binars.axiom_checkers import (
    associative,
    idempotent,
    is_left_identity,
    is_right_identity,
    left_distributive,
    right_distributive,
)


class BooleanRing(AbelianGroup):
    """
    a Boolean ring structure

    >>> add = {"0": {"0": "0", "1": "1"}, "1": {"0": "1", "1": "0"}}
    >>> neg = {"0": "0", "1": "1"}
    >>> mult = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}
    >>> ring = BooleanRing("test", {"add": add, "neg": neg, "mult": mult})
    >>> print(ring.mace4_format)
    0 + 0 = 0.
    0 + 1 = 1.
    1 + 0 = 1.
    1 + 1 = 0.
    neg(0) = 0.
    neg(1) = 1.
    0 * 0 = 0.
    0 * 1 = 0.
    1 * 0 = 0.
    1 * 1 = 1.
    <BLANKLINE>
    """

    def check_axioms(self) -> None:
        super().check_axioms()
        assert associative(self.operations["mult"])
        assert is_left_identity(self.operations["mult"], "1")
        assert is_right_identity(self.operations["mult"], "1")
        assert left_distributive(
            self.operations["mult"], self.operations["add"]
        )
        assert right_distributive(
            self.operations["mult"], self.operations["add"]
        )
        assert idempotent(self.operations["mult"])

    @property
    def operation_map(self) -> Dict[str, str]:
        return {"add": "+", "mult": "*"}



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Constants
==========
"""
FOR_X_Y = "\\<forall> x::finite_type. \\<forall> y::finite_type."
FOR_X_Y_Z = f"{FOR_X_Y} \\<forall> z::finite_type."
COMMUTATIVITY = f"({FOR_X_Y} f(x, y) = f(y, x))"
ASSOCIATIVITY = f"({FOR_X_Y_Z} f(x, f(y, z)) = f(f(x, y), z))"
IDEMPOTENCE = "(\\<forall> x::finite_type. f(x, x) = x)"
LEFT_IDENTITY = "(\\<forall> x::finite_type. f(C1, x) = x)"
RIGHT_IDENTITY = "(\\<forall> x::finite_type. f(x, C1) = x)"
LEFT_ZERO = "(\\<forall> x::finite_type. f(C0, x) = C0)"
RIGHT_ZERO = "(\\<forall> x::finite_type. f(x, C0) = C0)"
LEFT_DISTRIBUTIVITY = f"({FOR_X_Y_Z} f(x, g(y, z)) = g(f(x, y), f(x, z)))"
RIGHT_DISTRIBUTIVITY = f"({FOR_X_Y_Z} f(g(x, y), z) = g(f(x, z), f(y, z)))"
LEFT_ANTI_DISTRIBUTIVITY = f"({FOR_X_Y_Z} f(x, g(y, z)) = h(f(x, y), f(x, z)))"
RIGHT_ANTI_DISTRIBUTIVITY = (
    f"({FOR_X_Y_Z} f(g(x, y), z) = h(f(x, z), f(y, z)))"
)
PROJECTION = "(\\<forall> x::finite_type. f(f(x)) = x)"
DE_MORGAN = f"({FOR_X_Y} f(g(x, y)) = h(f(x), f(y)))"
ABSORPTION = f"({FOR_X_Y} f(x, g(x, y)) = x)"

LATTICE = [
    COMMUTATIVITY.replace("f(", "meet("),
    COMMUTATIVITY.replace("f(", "join("),
    ASSOCIATIVITY.replace("f(", "meet("),
    ASSOCIATIVITY.replace("f(", "join("),
    ABSORPTION.replace("f(", "meet(").replace("g(", "join("),
    ABSORPTION.replace("g(", "meet(").replace("f(", "join("),
]

RESIDUATED_BINAR = LATTICE + [
    LEFT_DISTRIBUTIVITY.replace("f(", "mult(").replace("g(", "join("),
    RIGHT_DISTRIBUTIVITY.replace("f(", "mult(").replace("g(", "join("),
    f"({FOR_X_Y_Z} meet(x, over(join(mult(x, y), z), y)) = x)",
    f"({FOR_X_Y_Z} meet(y, undr(x, join(mult(x, y), z))) = y)",
    f"({FOR_X_Y_Z} join(mult(over(x, y), y), x) = x)",
    f"({FOR_X_Y_Z} join(mult(y, undr(y, x)), x) = x)",
]

TRIVIAL_DISTRIBUTIVITY_LAWS = [
    LEFT_DISTRIBUTIVITY.replace("f(", "mult(").replace("g(", "join("),
    RIGHT_DISTRIBUTIVITY.replace("f(", "mult(").replace("g(", "join("),
    LEFT_DISTRIBUTIVITY.replace("f(", "undr(").replace("g(", "meet("),
    RIGHT_DISTRIBUTIVITY.replace("f(", "over(").replace("g(", "meet("),
    LEFT_ANTI_DISTRIBUTIVITY.replace("f(", "over(")
    .replace("g(", "join(")
    .replace("h(", "meet("),
    RIGHT_ANTI_DISTRIBUTIVITY.replace("f(", "undr(")
    .replace("g(", "join(")
    .replace("h(", "meet("),
]

BOUNDED_LATTICE = [
    LEFT_ZERO.replace("f(", "meet("),
    LEFT_IDENTITY.replace("f(", "meet("),
    RIGHT_ZERO.replace("f(", "meet("),
    RIGHT_IDENTITY.replace("f(", "meet("),
    LEFT_ZERO.replace("f(", "join(").replace("C0", "C1"),
    LEFT_IDENTITY.replace("f(", "join(").replace("C1", "C0"),
    RIGHT_ZERO.replace("f(", "join(").replace("C0", "C1"),
    RIGHT_IDENTITY.replace("f(", "join(").replace("C1", "C0"),
]

ORTHOCOMPLEMENTATION = [
    "(\\<forall> x::finite_type. meet(invo(x), x) = C0)",
    "(\\<forall> x::finite_type. join(invo(x), x) = C1)",
]

NON_TRIVIAL_DISTRIBUTIVITY_LAWS = [
    LEFT_DISTRIBUTIVITY.replace("f(", "mult(").replace("g(", "meet("),
    RIGHT_DISTRIBUTIVITY.replace("f(", "mult(").replace("g(", "meet("),
    LEFT_DISTRIBUTIVITY.replace("f(", "undr(").replace("g(", "join("),
    RIGHT_DISTRIBUTIVITY.replace("f(", "over(").replace("g(", "join("),
    LEFT_ANTI_DISTRIBUTIVITY.replace("f(", "over(")
    .replace("g(", "meet(")
    .replace("h(", "join("),
    RIGHT_ANTI_DISTRIBUTIVITY.replace("f(", "undr(")
    .replace("g(", "meet(")
    .replace("h(", "join("),
]

MODULARITY = (
    f"({FOR_X_Y_Z} f(g(x, z), g(y, z)) = g(f(g(x, z), y), z))".replace(
        "f(", "join("
    ).replace("g(", "meet(")
)

INVOLUTION = [
    DE_MORGAN.replace("f(", "invo(")
    .replace("g(", "join(")
    .replace("h(", "meet("),
    DE_MORGAN.replace("f(", "invo(")
    .replace("h(", "join(")
    .replace("g(", "meet("),
    PROJECTION.replace("f(", "invo("),
]

LEFT_INVERSE = "(\\<forall> x::finite_type. f(g(x), x) = C1)"
RIGHT_INVERSE = "(\\<forall> x::finite_type. f(x, g(x)) = C1)"
# see Definition 10 from https://doi.org/10.1155/2014/854168
PSEUDO_R0_ALGEBRA = (
    "("
    + " & ".join(
        [
            f"({FOR_X_Y} imp1(x, y) = imp2(inv1(y), inv1(x)))",
            f"({FOR_X_Y} imp2(x, y) = imp1(inv2(y), inv2(x)))",
        ]
    )
    + ")",
    "("
    + " & ".join(
        [
            LEFT_IDENTITY.replace("f(", "imp1("),
            LEFT_IDENTITY.replace("f(", "imp2("),
        ]
    )
    + ")",
    "("
    + " & ".join(
        [
            f"({FOR_X_Y_Z} meet(imp1(x, y), imp1(imp1(z, x), imp1(z, y)))"
            + " = imp1(x, y))",
            f"({FOR_X_Y_Z} meet(imp2(x, y), imp2(imp2(z, x), imp2(z, y)))"
            + " = imp2(x, y))",
        ]
    )
    + ")",
    "("
    + " & ".join(
        [
            LEFT_DISTRIBUTIVITY.replace("f(", "imp1(").replace("g(", "join("),
            LEFT_DISTRIBUTIVITY.replace("f(", "imp2(").replace("g(", "join("),
        ]
    )
    + ")",
    "("
    + " & ".join(
        [
            f"({FOR_X_Y} join(imp1(x, y), imp2(imp1(x, y), join(inv1(x), y)))"
            + " = C1)",
            f"({FOR_X_Y} join(imp2(x, y), imp1(imp2(x, y), join(inv2(x), y)))"
            + " = C1)",
        ]
    )
    + ")",
)



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Use Nitpick
============

A wrapper ‘do all’ script.

-  runs ``generate_theories.py`` which creates a new ``hyp2`` folder
   with initial hypotheses templates
-  then for each cardinality from 2 to 100 (hard-coded)
-  runs ``add_task.py`` which creates a ``task[n]`` folder for a
   particular cardinality with a respective task for ``Nitpick`` added
   to the templates in ``hyp[n]``
-  runs ``check_assumptions.py`` on a ``task[n]`` folder
-  runs ``filter_theories.py`` which filter theories with no
   counter-examples found to a new folder ``hyp[n+1]``
-  if the ``hyp[n+1]`` folder is empty, the script stops (that means
   counter-examples were found for all original hypotheses)

"""
import os
from typing import List, Optional

from residuated_binars.add_task import TaskType, add_task
from residuated_binars.check_assumptions import check_assumptions
from residuated_binars.filter_theories import filter_theories
from residuated_binars.generate_theories import independence_check


def use_nitpick(
    max_cardinality: int,
    independent_assumptions: List[str],
    additional_assumptions: List[str],
    check_subset_independence: bool,
    server_info: Optional[str] = None,
) -> None:
    """
    incrementally search for finite counter-examples

    :param max_cardinality: maximal cardinality of a model to search for
    :param independent_assumptions: a list of assumption which independence
        we want to check
    :param additional_assumptions: a list of additional assumptions
    :param check_subset_independence: whether to check every assumption from
        the list against all the rest or against any combination of the rest
    :param server_info: an info string of an Isabelle server
    :returns:
    """
    cardinality = 2
    hypotheses = f"hyp{cardinality}"
    independence_check(
        hypotheses,
        independent_assumptions,
        additional_assumptions,
        check_subset_independence,
    )
    while cardinality <= max_cardinality and os.listdir(hypotheses) != []:
        tasks = f"task{cardinality}"
        add_task(hypotheses, tasks, TaskType.NITPICK, cardinality)
        check_assumptions(tasks, server_info)
        cardinality += 1
        hypotheses = f"hyp{cardinality}"
        filter_theories(tasks, hypotheses)



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Filter Theories
================

-  reads from ``isabelle.out`` file from the input directory (this
   directory should be an output of ``check_assumption.py`` script)
-  filters only those theory files, for which neither finite model was
   found, nor the proof (depending on the task type)
-  copies filtered theory files from the input directory to another
   given directory

"""
import json
import os
import re
import shutil


def filter_theories(source_path: str, target_path: str) -> None:
    """
    get theory files from an existing folder and copy to another existing
    folder ones those of them, for which neither have a finite counter-example
    nor a proof

    :param source_path: where to look for processed theory files; should
        include an ``isabelle.out`` file with server's output
    :param target_path: where to put theory files without proofs or
        counter-examples
    :returns:
    """
    if not os.path.exists(target_path):
        os.mkdir(target_path)
    with open(
        os.path.join(source_path, "isabelle.out"), "r", encoding="utf-8"
    ) as out_file:
        final_line = [
            re.compile(".*FINISHED (.*)\n?").match(line)
            for line in out_file.readlines()
            if "FINISHED" in line
            and ("Sledgehammering" in line or "Nitpick" in line)
        ][0]
    if final_line is None:
        raise ValueError(f"Unexpected Isabelle server response: {final_line}")
    results = {
        node["theory_name"][6:]: [
            message["message"]
            for message in node["messages"]
            if any(
                nitpick_message in message["message"]
                for nitpick_message in (
                    "Try this: ",
                    "Timed out",
                    "Nitpick found a potentially spurious counterexample",
                    "Nitpick found a counterexample",
                    "Nitpick found no counterexample",
                )
            )
        ][0]
        for node in json.loads(final_line.group(1))["nodes"]
    }
    for result in results:
        if (
            "Nitpick found no counterexample" in results[result]
            or "Timed out" in results[result]
        ):
            shutil.copy(
                os.path.join(source_path, result + ".thy"),
                os.path.join(target_path, result + ".thy"),
            )



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Generate Theories
==================

This script creates a folder with a specified name and fills it with valid
Isabelle theory files. Each theory file contains only one lemma without a proof
and does not depend on any other theories. The exact statement of lemmas
is hard-coded. It may have the following assumptions:

-  lattice axioms
-  residuation axioms
-  existence of the latest and the greatest elements of the lattice
   (they always exist in finite models)
-  a definition of the involution operation
-  some combination of abstract distributivity laws (from
   ``ASSUMPTIONS`` variable of ``constants.py``)

The consequent of the lemma is one of the laws from ``ASSUMPTIONS``
which is not in the antecedent. So, if one has six laws in
``ASSUMPTIONS`` list, there will be ``6 * (2 ^ 5 - 1) = 186`` original
hypotheses to check for models of different cardinalities.

The theory files are called ``T[number].thy`` where ``number``
enumerates theory files starting from zero.
"""
import os
from itertools import combinations
from typing import List, Optional


def generate_isabelle_theory_file(
    theory_name: str, assumptions: List[str], goal: Optional[str] = None
) -> List[str]:
    """
    generate a text of Isabelle theory file with only ones lemma inside

    :param theory_name: name of a theory file
    :param assumptions: a list of lemma assumptions in Isabelle language
    :param goal: the lemma goal in Isabelle language
    :returns: a list of lines of a theory file
    """
    theory_text = [f"theory {theory_name}"]
    theory_text += [
        "imports Main",
        "begin",
        "datatype finite_type = finite_type_constants",
        'lemma "(',
    ]
    theory_text += [" &\n".join(assumptions)]
    if goal is not None:
        theory_text += [") \\<longrightarrow>"]
        theory_text += [goal]
    else:
        theory_text += [")"]
    theory_text += ['"', "oops", "end"]
    return theory_text


def independence_case(
    path: str,
    independent_assumptions: List[str],
    assumption_indices: List[int],
    goal_index: int,
    additional_assumptions: List[str],
) -> None:
    """
    generate a text of ``isabelle`` theory file for checking independence of
    one chosen assumpion from some subset of the rest

    :param path: a folder for storing theory files
    :param independent_assumptions: a list of assumption which independence
        we want to check
    :param assumption_indices: indices of assumption to use
    :param goal_index: index of a goal to prove
    :param additional_assumptions: a list of additional assumptions about
        the binars like the lattice reduct distributivity, existence of
        an involution operation, and multiplication associativity
    :returns:
    """
    assumption_list = list(assumption_indices)
    theory_name = f"T{''.join(map(str, assumption_list))}_{goal_index}"
    all_assumptions = [
        independent_assumptions[k] for k in assumption_indices
    ] + additional_assumptions
    theory_text = generate_isabelle_theory_file(
        theory_name,
        all_assumptions,
        independent_assumptions[goal_index],
    )
    with open(
        os.path.join(path, f"{theory_name}.thy"),
        "w",
        encoding="utf-8",
    ) as theory_file:
        theory_file.write("\n".join(theory_text))


def independence_check(
    path: str,
    independent_assumptions: List[str],
    additional_assumptions: List[str],
    check_subset_independence: bool,
) -> None:
    """
    generate a bunch of theory files to check independence of some assumptions
    given additional ones

    :param path: a folder for storing theory files
    :param independent_assumptions: a list of assumption which independence
        we want to check
    :param additional_assumptions: a list of additional assumptions
    :param check_subset_independence: whether to check every assumption from
        the list against all the rest or against any combination of the rest
    :returns:
    """
    if not os.path.exists(path):
        os.mkdir(path)
    total_assumptions_count = len(independent_assumptions)
    for goal_index in range(total_assumptions_count):
        for assumptions_count in range(1, total_assumptions_count):
            indices = tuple(
                list(range(0, goal_index))
                + list(range(goal_index + 1, total_assumptions_count))
            )
            if check_subset_independence:
                index_combinations = list(
                    combinations(
                        indices,
                        assumptions_count,
                    )
                )
            else:
                index_combinations = [indices]
            for assumption_indices in index_combinations:
                independence_case(
                    path,
                    independent_assumptions,
                    list(assumption_indices),
                    goal_index,
                    additional_assumptions,
                )



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Script for adding a task
=========================

-  for every theory file in a given folder
-  creates a new file with the same name in another given folder
-  the output file includes the same one lemma as the input
-  but the ‘task’ is changed according to the script’s parameters

Possible task types:

- ``TaskType.NITPICK``, demands a ``cardinality`` to be provided as well
  (finite model search)
- ``TaskType.SLEDGEHAMMER`` (automated proof search)

Both task types have a hard-coded timeout of ``1000000`` seconds.

"""
import os
import re
from enum import Enum


class TaskType(Enum):
    """
    type of tasks to ask ``isabelle`` server to perform
    """

    SLEDGEHAMMER = "sledgehammer[timeout=1000000]"
    NITPICK = "nitpick[timeout=1000000,max_threads=0]"


def add_task(
    source_path: str,
    target_path: str,
    task_type: TaskType,
    cardinality: int = 1,
) -> None:
    """
    take theory files from an existing folder and change tasks in them to given
    ones

    :param source_path: a directory where to get theory files to add tasks to
    :param target_path: where to put new theory files with added tasks
    :param task_type: use Nitpick or Sledgehammer (disprove by finding a finite
        coutner-example or prove)
    :param cardinality: a cardinality of finite model to find (only for Nitpick
        tasks)
    :returns:
    """
    if not os.path.exists(target_path):
        os.mkdir(target_path)
    for theory_name in os.listdir(source_path):
        with open(
            os.path.join(source_path, theory_name), "r", encoding="utf-8"
        ) as theory_file:
            theory_text = theory_file.read()
        theory_text = re.sub(
            "datatype finite_type =.*\n",
            "datatype finite_type = "
            + " | ".join(map(lambda i: "C" + str(i), range(cardinality)))
            + "\n",
            re.sub(
                '"\n.*\n?oops',
                '"\n' + task_type.value + "\noops",
                theory_text,
            ),
        )
        with open(
            os.path.join(target_path, theory_name), "w", encoding="utf-8"
        ) as theory_file:
            theory_file.write(theory_text)



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Abelian Group
==============
"""
from typing import Dict

from residuated_binars.algebraic_structure import AlgebraicStructure
from residuated_binars.axiom_checkers import (
    associative,
    commutative,
    is_left_identity,
    is_left_inverse,
    is_right_identity,
    is_right_inverse,
)


class AbelianGroup(AlgebraicStructure):
    """
    an Abelian group structure

    >>> add = {"0": {"0": "0"}}
    >>> group = AbelianGroup("test", {"add": add, "neg": {"0": "0"}})
    >>> print(group.mace4_format)
    0 + 0 = 0.
    neg(0) = 0.
    <BLANKLINE>
    """

    def check_axioms(self) -> None:
        assert associative(self.operations["add"])
        assert commutative(self.operations["add"])
        assert is_left_identity(self.operations["add"], "0")
        assert is_right_identity(self.operations["add"], "0")
        assert is_left_inverse(
            self.operations["add"], self.operations["neg"], "0"
        )
        assert is_right_inverse(
            self.operations["add"], self.operations["neg"], "0"
        )

    @property
    def operation_map(self) -> Dict[str, str]:
        return {"add": "+"}



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Axiom Checkers
===============
"""
from typing import Dict

from residuated_binars.algebraic_structure import CayleyTable


def associative(cayley_table: CayleyTable) -> bool:
    """
    >>> associative({"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "0"}})
    True
    >>> associative({"0": {"0": "1", "1": "0"}, "1": {"0": "0", "1": "0"}})
    False

    :param cayley_table: a multiplication table of a binary operation
    :returns: whether the operation is associative or not
    """
    for one in cayley_table.keys():
        for two in cayley_table.keys():
            for three in cayley_table.keys():
                if (
                    cayley_table[one][cayley_table[two][three]]
                    != cayley_table[cayley_table[one][two]][three]
                ):
                    return False
    return True


def is_left_identity(cayley_table: CayleyTable, identity: str) -> bool:
    """
    >>> is_left_identity(
    ...     {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}, "1"
    ... )
    True
    >>> is_left_identity(
    ...     {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}, "0"
    ... )
    False

    :param cayley_table: a multiplication table of a binary operation
    :param identity: a symbol for identity
    :returns: whether ``identity`` is a left identity for a Cayley table
    """
    for one in cayley_table.keys():
        if cayley_table[identity][one] != one:
            return False
    return True


def is_right_identity(cayley_table: CayleyTable, identity: str) -> bool:
    """
    >>> is_right_identity(
    ...     {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}, "1"
    ... )
    True
    >>> is_right_identity(
    ...     {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}, "0"
    ... )
    False

    :param cayley_table: a multiplication table of a binary operation
    :param identity: a symbol for identity
    :returns: whether ``identity`` is a right identity for a Cayley table
    """
    for one in cayley_table.keys():
        if cayley_table[one][identity] != one:
            return False
    return True


def is_left_inverse(
    cayley_table: CayleyTable, inverse: Dict[str, str], identity: str
) -> bool:
    """
    >>> op = {"0": {"0": "1", "1": "1"}, "1": {"0": "1", "1": "1"}}
    >>> inv = {"0": "1", "1": "1"}
    >>> is_left_inverse(op, inv, "1")
    True
    >>> is_left_inverse(op, inv, "0")
    False

    :param cayley_table: a multiplication table of a binary operation
    :param inverse: a map for the operation of inversion
    :param identity: a symbol for identity
    :returns: whether ``inverse`` is a left inverse for a Cayley table
    """
    for one in cayley_table.keys():
        if cayley_table[inverse[one]][one] != identity:
            return False
    return True


def is_right_inverse(
    cayley_table: CayleyTable, inverse: Dict[str, str], identity: str
) -> bool:
    """
    >>> op = {"0": {"0": "1", "1": "1"}, "1": {"0": "1", "1": "1"}}
    >>> inv = {"0": "1", "1": "1"}
    >>> is_right_inverse(op, inv, "1")
    True
    >>> is_right_inverse(op, inv, "0")
    False

    :param cayley_table: a multiplication table of a binary operation
    :param inverse: a map for the operation of inversion
    :param identity: a symbol for identity
    :returns: whether ``inverse`` is a right inverse for a Cayley table
    """
    for one in cayley_table.keys():
        if cayley_table[one][inverse[one]] != identity:
            return False
    return True


def commutative(cayley_table: CayleyTable) -> bool:
    """
    >>> commutative({"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "0"}})
    True
    >>> commutative({"0": {"0": "0", "1": "1"}, "1": {"0": "0", "1": "0"}})
    False

    :param cayley_table: a multiplication table of a binary operation
    :returns: whether the operation is commutative or not
    """
    for one in cayley_table.keys():
        for two in cayley_table.keys():
            if cayley_table[one][two] != cayley_table[two][one]:
                return False
    return True


def idempotent(cayley_table: CayleyTable) -> bool:
    """
    >>> idempotent({"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}})
    True
    >>> idempotent({"0": {"0": "0", "1": "1"}, "1": {"0": "0", "1": "0"}})
    False

    :param cayley_table: a multiplication table of a binary operation
    :returns: whether the operation is idempotent or not
    """
    for one in cayley_table.keys():
        if cayley_table[one][one] != one:
            return False
    return True


def left_distributive(table1: CayleyTable, table2: CayleyTable) -> bool:
    """
    >>> operation = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "0"}}
    >>> left_distributive(operation, operation)
    True
    >>> operation = {"0": {"0": "1", "1": "0"}, "1": {"0": "0", "1": "0"}}
    >>> left_distributive(operation, operation)
    False

    :param table1: a multiplication table of a binary operation
    :param table2: a multiplication table of another binary operation
    :returns: whether the first operation is left distributive with respect to
        the second one or not
    """
    for one in table1.keys():
        for two in table1.keys():
            for three in table1.keys():
                if (
                    table1[one][table2[two][three]]
                    != table2[table1[one][two]][table1[one][three]]
                ):
                    return False
    return True


def right_distributive(table1: CayleyTable, table2: CayleyTable) -> bool:
    """
    >>> operation = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "0"}}
    >>> right_distributive(operation, operation)
    True
    >>> operation = {"0": {"0": "0", "1": "1"}, "1": {"0": "0", "1": "0"}}
    >>> right_distributive(operation, operation)
    False

    :param table1: a multiplication table of a binary operation
    :param table2: a multiplication table of another binary operation
    :returns: whether the first operation is right distributive with respect to
        the second one or not
    """
    for one in table1.keys():
        for two in table1.keys():
            for three in table1.keys():
                if (
                    table1[table2[one][two]][three]
                    != table2[table1[one][three]][table1[two][three]]
                ):
                    return False
    return True


def absorbs(table1: CayleyTable, table2: CayleyTable) -> bool:
    """
    >>> conjuction = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}
    >>> disjunction = {"0": {"0": "0", "1": "1"}, "1": {"0": "1", "1": "1"}}
    >>> absorbs(conjuction, disjunction)
    True
    >>> absorbs(conjuction, conjuction)
    False

    :param table1: a multiplication table of a binary operation
    :param table2: a multiplication table of another binary operation
    :returns: whether the absorption law is true with respect to
        two binary operation
    """
    for one in table1.keys():
        for two in table1.keys():
            if table1[one][table2[one][two]] != one:
                return False
    return True


def is_left_zero(cayley_table: CayleyTable, zero: str) -> bool:
    """
    >>> table = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}
    >>> is_left_zero(table, "0")
    True
    >>> is_left_zero(table, "1")
    False

    :param cayley_table: a multiplication table of a binary operation
    :param zero: a symbol for the zero
    :returns: whether ``zero`` is a left zero for a Cayley table
    """
    for one in cayley_table.keys():
        if cayley_table[zero][one] != zero:
            return False
    return True


def is_right_zero(cayley_table: CayleyTable, zero: str) -> bool:
    """
    >>> table = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}
    >>> is_right_zero(table, "0")
    True
    >>> is_right_zero(table, "1")
    False

    :param cayley_table: a multiplication table of a binary operation
    :param zero: a symbol for the zero
    :returns: whether ``zero`` is a right zero for a Cayley table
    """
    for one in cayley_table.keys():
        if cayley_table[one][zero] != zero:
            return False
    return True



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Check Assumptions
==================

-  gets all theory files from a given directory
-  constructs a command for Isabelle server to process these files
-  saves the log of Isabelle server replies to the file named
   ``isabelle.out`` in the directory where the theory files are

This script depends on `Python client for Isabelle
server <https://pypi.org/project/isabelle-client>`__.

"""
import logging
import os
import sys
from typing import Optional

import nest_asyncio
from isabelle_client import get_isabelle_client
from isabelle_client.utils import start_isabelle_server


def get_customised_logger(task_folder: str) -> logging.Logger:
    """
    get a nice logger

    :param rask_folder: a base folder (and a task name)
    """
    logfile_name = os.path.join(task_folder, "isabelle.out")
    if os.path.exists(logfile_name):
        os.remove(logfile_name)
    logger = logging.getLogger(os.path.basename(task_folder))
    handler = logging.FileHandler(logfile_name)
    handler.setFormatter(logging.Formatter("%(asctime)s: %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return logger


def check_assumptions(path: str, server_info: Optional[str] = None) -> None:
    """
    ask Isabelle server to process all theory files in a given path

    :param path: a folder with theory files
    :param server_info: an info string of an Isabelle server
    :returns:
    """
    nest_asyncio.apply()
    theories = [
        theory_name[0]
        for theory_name in [
            os.path.splitext(theory_file) for theory_file in os.listdir(path)
        ]
        if theory_name[1] == ".thy"
    ]
    new_server_info = _start_server_if_needed(path, server_info)
    isabelle_client = get_isabelle_client(new_server_info)
    isabelle_client.logger = get_customised_logger(path)
    isabelle_client.use_theories(
        theories=theories, master_dir=get_abs_path(path), watchdog_timeout=0
    )
    if server_info is None:
        isabelle_client.shutdown()


def get_abs_path(path: str) -> str:
    """
    :param path: a path
    :returns: an absolute path, corrected to CygWin path for Windows
    """
    abs_path = os.path.abspath(path)
    if sys.platform == "win32":
        abs_path = abs_path.replace("C:\\", "/cygdrive/c/").replace("\\", "/")
    return abs_path


def _start_server_if_needed(path: str, server_info: Optional[str]) -> str:
    if server_info is None:
        new_server_info, _ = start_isabelle_server(
            log_file=os.path.join(path, "server.log"),
            name=os.path.basename(path),
        )
    else:
        new_server_info = server_info
    return new_server_info



# Copyright 2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Bounded Lattice
===============
"""
from residuated_binars.algebraic_structure import BOT, TOP
from residuated_binars.axiom_checkers import (
    is_left_identity,
    is_left_zero,
    is_right_identity,
    is_right_zero,
)
from residuated_binars.lattice import Lattice


class BoundedLattice(Lattice):
    r"""
        a representation of a bounded lattice

    >>> join = {"0": {"0": "0", "1": "1"}, "1": {"0": "1", "1": "1"}}
    >>> meet = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}
    >>> lattice = Lattice("test", {"join": join, "meet": meet})
    >>> lattice.canonise_symbols()
    >>> BoundedLattice(lattice.label, lattice.operations)
    {'join': [[0, 1], [1, 1]], 'meet': [[0, 0], [0, 1]]}
    """

    def check_axioms(self) -> None:
        super().check_axioms()
        assert is_left_identity(self.operations["meet"], TOP)
        assert is_right_identity(self.operations["meet"], TOP)
        assert is_left_identity(self.operations["join"], BOT)
        assert is_right_identity(self.operations["join"], BOT)
        assert is_left_zero(self.operations["join"], TOP)
        assert is_right_zero(self.operations["join"], TOP)
        assert is_left_zero(self.operations["meet"], BOT)
        assert is_right_zero(self.operations["meet"], BOT)



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Residuated Binar
=================
"""
from typing import Dict

from residuated_binars.axiom_checkers import (
    left_distributive,
    right_distributive,
)
from residuated_binars.lattice import BOT, Lattice


class ResiduatedBinar(Lattice):
    r"""
        a representation of a residuated binar (with involution)

    >>> join = {"0": {"0": "0", "1": "1"}, "1": {"0": "1", "1": "1"}}
    >>> meet = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}
    >>> mult = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "0"}}
    >>> const = {"0": {"0": "1", "1": "1"}, "1": {"0": "1", "1": "1"}}
    >>> binar = ResiduatedBinar(
    ...     label="test",
    ...     operations={
    ...         "join": join, "meet": meet, "mult": mult, "over": const,
    ...         "undr": const, "invo": {"0": "1", "1": "0"}
    ...     }
    ... )
    >>> print(binar.latex_mult_table)
    \begin{table}[]
    \begin{tabular}{l|ll}
    $\cdot$ & $0$ & $1$\\\hline
    $0$ & $0$ & $0$ & \\
    $1$ & $0$ & $0$ & \\
    \end{tabular}
    \end{table}
    <BLANKLINE>
    >>> print(binar.markdown_mult_table)
    |*|0|1|
    |-|-|-|
    |**0**|0|0|
    |**1**|0|0|
    <BLANKLINE>
    >>> mult["0"]["0"] = "1"
    >>> ResiduatedBinar("test", {"join": join, "meet": meet, "mult": mult,
    ...     "over": const, "undr": const})
    Traceback (most recent call last):
     ...
    ValueError: multiplication must be distributive over join
    >>> mult["0"]["0"] = "0"
    >>> const["0"]["0"] = "0"
    >>> ResiduatedBinar("test", {"join": join, "meet": meet, "mult": mult,
    ...     "over": const, "undr": const})
    Traceback (most recent call last):
     ...
    ValueError: check residuated binars axioms!
    >>> mult = {"0": {"0": "0", "1": "0"}, "1": {"0": "1", "1": "1"}}
    >>> undr = {"0": {"0": "1", "1": "1"}, "1": {"0": "0", "1": "1"}}
    >>> ResiduatedBinar("test", {"join": join, "meet": meet, "mult": mult,
    ...     "over": mult, "undr": undr})
    Traceback (most recent call last):
     ...
    ValueError: check residuated binars axioms!
    >>> print(binar.mace4_format[:10])
    0 v 0 = 0.
    """

    def check_axioms(self) -> None:
        super().check_axioms()
        if not left_distributive(
            self.operations["mult"], self.operations["join"]
        ) or not right_distributive(
            self.operations["mult"], self.operations["join"]
        ):
            raise ValueError("multiplication must be distributive over join")
        if not self._check_residuated_binars_axioms():
            raise ValueError("check residuated binars axioms!")

    def _check_residuated_binars_axioms(self) -> bool:
        for one in self.symbols:
            for two in self.symbols:
                if (
                    self.operations["join"][
                        self.operations["mult"][
                            self.operations["over"][one][two]
                        ][two]
                    ][one]
                    != one
                    or self.operations["join"][
                        self.operations["mult"][two][
                            self.operations["undr"][two][one]
                        ]
                    ][one]
                    != one
                ):
                    return False
                for three in self.symbols:
                    if (
                        self.operations["meet"][one][
                            self.operations["over"][
                                self.operations["join"][
                                    self.operations["mult"][one][two]
                                ][three]
                            ][two]
                        ]
                        != one
                        or self.operations["meet"][two][
                            self.operations["undr"][one][
                                self.operations["join"][
                                    self.operations["mult"][one][two]
                                ][three]
                            ]
                        ]
                        != two
                    ):
                        return False
        return True

    @property
    def operation_map(self) -> Dict[str, str]:
        res = super().operation_map
        res.update({"over": "/", "undr": "\\", "mult": "*"})
        return res

    @property
    def latex_mult_table(self) -> str:
        """
        :returns: a LaTeX representation of a multiplication table
        """
        table = (
            "\\begin{table}[]\n"
            + "\\begin{tabular}"
            + f"{{l|{''.join((self.cardinality) * 'l')}}}\n"
            + "$"
            + "$ & $".join([r"\cdot"] + self.symbols)
            + "$\\\\\\hline\n"
        )
        for row in self.symbols:
            table += "$" + row + "$ & "
            for col in self.symbols:
                table += "$" + self.operations["mult"][row][col] + "$"
                if col != BOT:
                    table += " & "
            if row != BOT:
                table += r"\\"
            table += "\n"
        table += "\\end{tabular}\n" + "\\end{table}\n"
        return table

    @property
    def markdown_mult_table(self) -> str:
        """
        :returns: a Markdown representation of a multiplication table
        """
        table = "|*|" + "|".join(self.symbols) + "|\n"
        table += "|" + (1 + len(self.symbols)) * "-|" + "\n"
        for row in self.symbols:
            table += "|**" + row + "**|"
            for col in self.symbols:
                table += self.operations["mult"][row][col] + "|"
            table += "\n"
        return table



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Parser
=======
"""
import json
import re
from typing import Any, Dict, List, Union

from residuated_binars.algebraic_structure import (
    AlgebraicStructure,
    CayleyTable,
)
from residuated_binars.lattice import Lattice
from residuated_binars.residuated_binar import ResiduatedBinar


def parse_binary_operation(line: str) -> CayleyTable:
    """
    parse text describing a binary operation in Isabelle server response

    :param line: a part of Isabelle server response, representing a binary
        operation
    :returns: a Cayley table
    """
    table: CayleyTable = {}
    regex = re.compile(
        r"\(([\w\\\^\<\>]+), ([\w\\\^\<\>]+)\) := ([\w\\\^\<\>]+)"
    )
    match = regex.search(line)
    while match is not None:
        pos = match.span()[-1] + 1
        args = list(match.groups())
        if args[0] not in table:
            table[args[0]] = {}
        table[args[0]][args[1]] = args[2]
        match = regex.search(line, pos)
    return table


def parse_unary_operation(line: str) -> Dict[str, str]:
    """
    parse text describing a unary operation in Isabelle server response

    :param line: a part of Isabelle server response, representing an unary
        operation
    :returns: an inner representation of an unary operation
    """
    table: Dict[str, str] = {}
    regex = re.compile(r"([\w\\\<\>\^]+) := ([\w\\\<\>\^]+)")
    match = regex.search(line)
    while match is not None:
        pos = match.span()[-1] + 1
        args = list(match.groups())
        table[args[0]] = args[1]
        match = regex.search(line, pos)
    return table


def choose_algebraic_structure(
    label: str, operations: Dict[str, Dict[str, Any]]
) -> AlgebraicStructure:
    """
    :param label: a name of that particular algebraic structure example
    :param operations: a dictionary of unary and binary operations
    :returns: an algebraic structure of a concrete type (
        depending on the signature)
    """
    sorted_ops = sorted(operations.keys())
    if sorted_ops == ["join", "meet"]:
        return Lattice(label, operations)
    if sorted_ops in [
        [
            "join",
            "meet",
            "mult",
            "over",
            "undr",
        ],
        ["invo", "join", "meet", "mult", "over", "undr"],
    ]:
        return ResiduatedBinar(label, operations)
    return AlgebraicStructure(label, operations)


def isabelle_format_to_algebra(
    isabelle_message: str, label: str
) -> AlgebraicStructure:
    """
    :param isabelle_message: a body of reply from Isabelle server (in JSON)
    :param label: a name of the theory for which we got a relply from server
    :returns: a residuated binar
    """
    regex = re.compile(
        r"    (\w+) =\n? +\(\\<lambda>x\. _\)\n? *\(([^\.]+)\)\n?",
        re.DOTALL,
    )
    match = regex.search(isabelle_message)
    operations: Dict[str, Union[CayleyTable, Dict[str, str]]] = {}
    while match is not None:
        table: Union[CayleyTable, Dict[str, str]] = parse_binary_operation(
            match.group(2)
        )
        if not table:
            table = parse_unary_operation(match.group(2))
        operations[match.group(1)] = table
        pos = match.span()[0] + 1
        match = regex.search(isabelle_message, pos)
    return choose_algebraic_structure(label, operations)


def isabelle_response_to_algebra(filename: str) -> List[AlgebraicStructure]:
    """
    read file with replies from ``isabelle`` server and parse all residuated
    binars from it

    >>> import sys
    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> import os
    >>> len(isabelle_response_to_algebra(
    ...     files("residuated_binars")
    ...     .joinpath(os.path.join("resources", "isabelle2.out"))
    ... ))
    6

    :param filename: a name of a file to which all replies from Isabelle server
        where written
    :returns: a list of algebraic structures
    """
    with open(filename, "r", encoding="utf-8") as isabelle_log:
        nodes = json.loads(
            [
                line
                for line in isabelle_log.readlines()
                if "FINISHED" in line and "lambda" in line
            ][0][9:]
        )["nodes"]
    messages = [
        (
            [
                message["message"]
                for message in node["messages"]
                if "lambda" in message["message"]
            ][:1],
            node["theory_name"].split(".")[1],
        )
        for node in nodes
    ]
    return [
        isabelle_format_to_algebra(message[0][0], message[1])
        for message in messages
        if message[0] != []
    ]



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Lattice
========
"""
from typing import Dict, List, Tuple

import graphviz

from residuated_binars.algebraic_structure import BOT, TOP, AlgebraicStructure
from residuated_binars.axiom_checkers import absorbs, associative, commutative


class Lattice(AlgebraicStructure):
    r"""
        a representation of a lattice

    >>> join = {"0": {"0": "0", "1": "1"}, "1": {"0": "1", "1": "1"}}
    >>> meet = {"0": {"0": "0", "1": "0"}, "1": {"0": "0", "1": "1"}}
    >>> lattice = Lattice("test", {"join": join, "meet": meet})
    >>> print(lattice.mace4_format)
    0 v 0 = 0.
    0 v 1 = 1.
    1 v 0 = 1.
    1 v 1 = 1.
    0 ^ 0 = 0.
    0 ^ 1 = 0.
    1 ^ 0 = 0.
    1 ^ 1 = 1.
    <BLANKLINE>
    >>> lattice.canonise_symbols()
    >>> print(lattice.graphviz_repr)
    graph {
        "⟙" -- "⟘"
    }
    >>> join["0"]["1"] = "0"
    >>> Lattice("test", {"join": join, "meet": meet})
    Traceback (most recent call last):
     ...
    ValueError: join is not commutative
    >>> join["0"]["1"] = "1"
    >>> meet["0"]["1"] = "1"
    >>> Lattice("test", {"join": join, "meet": meet})
    Traceback (most recent call last):
     ...
    ValueError: meet is not commutative
    >>> meet["0"]["1"] = "0"
    >>> join["0"]["0"] = "1"
    >>> Lattice("test", {"join": join, "meet": meet})
    Traceback (most recent call last):
     ...
    ValueError: absorption laws fail
    >>> join = {'0': {'0': '1', '1': '0'}, '1': {'0': '0', '1': '0'}}
    >>> meet = {'0': {'0': '0', '1': '0'}, '1': {'0': '0', '1': '1'}}
    >>> Lattice("test", {"join": join, "meet": meet})
    Traceback (most recent call last):
     ...
    ValueError: join is not associative
    >>> join = {'0': {'0': '0', '1': '0'}, '1': {'0': '0', '1': '1'}}
    >>> meet = {'0': {'0': '1', '1': '0'}, '1': {'0': '0', '1': '0'}}
    >>> Lattice("test", {"join": join, "meet": meet})
    Traceback (most recent call last):
     ...
    ValueError: meet is not associative
    """

    def check_axioms(self) -> None:
        self._check_commutativity_and_associativity()
        if not absorbs(
            self.operations["meet"], self.operations["join"]
        ) or not absorbs(self.operations["join"], self.operations["meet"]):
            raise ValueError("absorption laws fail")

    def _check_commutativity_and_associativity(self):
        if not commutative(self.operations["join"]):
            raise ValueError("join is not commutative")
        if not commutative(self.operations["meet"]):
            raise ValueError("meet is not commutative")
        if not associative(self.operations["join"]):
            raise ValueError("join is not associative")
        if not associative(self.operations["meet"]):
            raise ValueError("meet is not associative")

    @property
    def operation_map(self) -> Dict[str, str]:
        return {"meet": "^", "join": "v"}

    @property
    def more(self) -> Dict[str, List[str]]:
        """
        :returns: some representation of a 'more' relation of a lattice reduct
            of the lattice
        """
        relation: Dict[str, List[str]] = {}
        for one in self.symbols:
            relation[one] = []
            for two in self.symbols:
                if self.operations["meet"][one][two] == two and one != two:
                    relation[one].append(two)
        return relation

    @property
    def hasse(self) -> List[Tuple[str, str]]:
        """
        :returns: some representation of a Hasse diagram of a lattice reduct of
            the lattice
        """
        more = {
            pair[0]: pair[1]
            for pair in sorted(
                self.more.items(),
                key=lambda key_value: -len(key_value[1]),
            )
        }
        hasse = []
        for higher in more:
            nearest = set(more[higher])
            for lower in more[higher]:
                nearest = nearest.difference(set(more[lower]))
            hasse += [(higher, neighbour) for neighbour in list(nearest)]
        return hasse

    @property
    def graphviz_repr(self) -> str:
        """
        :returns: a representation usable by ``graphviz`` of a Hasse diagram of
            a lattice reduct of the lattice
        """
        graph = graphviz.Graph()
        for pair in self.hasse:
            graph.edge(pair[0], pair[1])
        return graph

    def canonise_symbols(self) -> None:
        """
        renumerate lattice's items in a canonical way
        """
        symbol_map = {
            pair[1]: pair[0]
            for pair in zip(
                [BOT]
                + [chr(ord("a") + i) for i in range(self.cardinality - 2)]
                + [TOP],
                [
                    key
                    for key, value in sorted(
                        self.more.items(),
                        key=lambda key_value: (
                            len(key_value[1]),
                            key_value[0],
                        ),
                    )
                ],
            )
        }
        self.remap_symbols(symbol_map)



# Copyright 2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Pseudo-:math:`R_0` Algebra
==========================
"""
from residuated_binars.algebraic_structure import TOP
from residuated_binars.pseudo_weak_r0_algebra import PseudoWeakR0Algebra


class PseudoR0Algebra(PseudoWeakR0Algebra):
    """
    a representation of a peudo-:math:`R_0` algebra

    for more info look `here <https://doi.org/10.1155/2014/854168>`__

    >>> from residuated_binars.algebraic_structure import BOT
    >>> imp = {
    ...     BOT: {BOT: TOP, TOP: TOP, "C3": TOP, "C2": TOP},
    ...     TOP: {BOT: BOT, TOP: TOP, "C3": "C3", "C2": "C2"},
    ...     "C3": {BOT: "C2", TOP: TOP, "C3": TOP, "C2": "C3"},
    ...     "C2": {BOT: "C3", TOP: TOP, "C3": TOP, "C2": TOP}
    ... }
    >>> inv = {BOT: TOP, TOP: BOT, "C3": "C2", "C2": "C3"}
    >>> join = {
    ...     BOT: {BOT: BOT, TOP: TOP, "C3": "C3", "C2": "C2"},
    ...     TOP: {BOT: TOP, TOP: TOP, "C3": TOP, "C2": TOP},
    ...     "C3": {BOT: "C3", TOP: TOP, "C3": "C3", "C2": "C3"},
    ...     "C2": {BOT: "C2", TOP: TOP, "C3": "C3", "C2": "C2"}
    ... }
    >>> meet = {
    ...     BOT: {BOT: BOT, TOP: BOT, "C3": BOT, "C2": BOT},
    ...     TOP: {BOT: BOT, TOP: TOP, "C3": "C3", "C2": "C2"},
    ...     "C3": {BOT: BOT, TOP: "C3", "C3": "C3", "C2": "C2"},
    ...     "C2": {BOT: BOT, TOP: "C2", "C3": "C2", "C2": "C2"}
    ... }
    >>> operations = {"imp1": imp, "imp2": imp, "inv1": inv, "inv2": inv,
    ...     "join": join, "meet": meet}
    >>> PseudoR0Algebra("no P5", operations)
    Traceback (most recent call last):
     ...
    ValueError: P5 axiom doesn't hold
    >>> imp = {
    ...     BOT: {BOT: TOP, TOP: TOP, "C3": TOP, "C2": TOP},
    ...     TOP: {BOT: BOT, TOP: TOP, "C3": "C3", "C2": "C2"},
    ...     "C2": {BOT: "C3", TOP: TOP, "C2": TOP, "C3": "C3"},
    ...     "C3": {BOT: "C2", TOP: TOP, "C2": "C2", "C3": TOP}
    ... }
    >>> operations = {"imp1": imp, "imp2": imp, "inv1": inv, "inv2": inv,
    ...     "join": join, "meet": meet}
    >>> PseudoR0Algebra("no P5", operations)
    Traceback (most recent call last):
     ...
    ValueError: P4 axiom doesn't hold
    >>> imp = {
    ...     BOT: {BOT: TOP, TOP: TOP, "C3": TOP, "C2": TOP},
    ...     TOP: {BOT: BOT, TOP: TOP, "C3": "C3", "C2": "C2"},
    ...     "C2": {BOT: "C2", TOP: TOP, "C2": TOP, "C3": TOP},
    ...     "C3": {BOT: "C3", TOP: TOP, "C2": TOP, "C3": TOP}
    ... }
    >>> inv = {BOT: TOP, TOP: BOT, "C3": "C3", "C2": "C2"}
    >>> operations = {"imp1": imp, "imp2": imp, "inv1": inv, "inv2": inv,
    ...     "join": join, "meet": meet}
    >>> PseudoR0Algebra("no P5", operations)
    Traceback (most recent call last):
     ...
    ValueError: P3 axiom doesn't hold
    """

    def check_axioms(self) -> None:
        super().check_axioms()
        try:
            for one in self.operations["meet"].keys():
                for two in self.operations["meet"].keys():
                    assert (
                        self.operations["join"][
                            self.operations["imp1"][one][two]
                        ][
                            self.operations["imp2"][
                                self.operations["imp1"][one][two]
                            ][
                                self.operations["join"][
                                    self.operations["inv1"][one]
                                ][two]
                            ]
                        ]
                        == TOP
                    )
                    assert (
                        self.operations["join"][
                            self.operations["imp2"][one][two]
                        ][
                            self.operations["imp1"][
                                self.operations["imp2"][one][two]
                            ][
                                self.operations["join"][
                                    self.operations["inv2"][one]
                                ][two]
                            ]
                        ]
                        == TOP
                    )
        except AssertionError as error:
            raise ValueError("P5 axiom doesn't hold") from error



# Copyright 2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Short scripts useful for other submodules
=========================================

"""
import os
import shutil
from glob import glob
from typing import Sequence


def remove_dirs(dir_list: Sequence[str]) -> None:
    """
    >>> os.path.exists("remove-test1")
    False
    >>> os.path.exists("remove-test2")
    False
    >>> os.mkdir("remove-test1")
    >>> os.mkdir(os.path.join("remove-test1", "some"))
    >>> os.mkdir("remove-test2")
    >>> os.path.exists(os.path.join("remove-test1", "some"))
    True
    >>> os.path.exists("remove-test2")
    True
    >>> remove_dirs(("remove-test*",))
    >>> os.path.exists("remove-test1")
    False
    >>> os.path.exists("remove-test2")
    False

    :param dir_list: subfolders of the current folder to remove completely
        (no questions asked!)
    :returns:
    """
    for some_dir in dir_list:
        for path in glob(os.path.join(".", some_dir)):
            shutil.rmtree(path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['residuated_binars']

package_data = \
{'': ['*'], 'residuated_binars': ['resources/*']}

install_requires = \
['graphviz', 'isabelle-client']

setup_kwargs = {
    'name': 'residuated-binars',
    'version': '0.0.1',
    'description': 'Package for generating and validating examples of different algebraic structures using Isabelle proof assistant',
    'long_description': '..\n  Copyright 2021-2022 Boris Shminke\n\n  Licensed under the Apache License, Version 2.0 (the "License");\n  you may not use this file except in compliance with the License.\n  You may obtain a copy of the License at\n\n      https://www.apache.org/licenses/LICENSE-2.0\n\n  Unless required by applicable law or agreed to in writing, software\n  distributed under the License is distributed on an "AS IS" BASIS,\n  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n  See the License for the specific language governing permissions and\n  limitations under the License.\n\n|Binder|\\ |PyPI version|\\ |CircleCI|\\ |codecov|\\ |RTFD|\n\nGenerating Algebraic Structures with Isabelle\n==============================================\n\n.. attention::\n   If you\'re looking for a reproducible example for AITP 2021 paper, find it `here <https://residuated-binars.readthedocs.io/en/latest/aitp2021.html>`__.\n\nThis package serves for generating and validating examples of different algebraic structures using `Isabelle proof assistant <https://isabelle.in.tum.de>`__.\n\n.. _how-to-install:\n\nDependencies\n=============\nMake sure that an installation of Isabelle is on the ``$PATH``\n\nHow to Install\n===============\n\nThe best way to install ``residuated-binars`` is to use ``pip``::\n  \n    pip install residuated-binars\n     \nAlternatively, one can use Docker:\n\n.. code:: sh\n\n      docker build -t residuated-binars https://github.com/inpefess/residuated-binars.git\n      docker run -it --rm -p 8888:8888 residuated-binars jupyter-lab --ip=0.0.0.0 --port=8888 --no-browser\n\nFinally, one can run it on\n`Binder <https://mybinder.org/v2/gh/inpefess/residuated-binars/HEAD?labpath=reproducing-residuated-binars-papers.ipynb>`__\n\n\nHow to Use\n===========\n\nSee ``examples/residuated-binars-example.ipynb``.\n\n.. |CircleCI| image:: https://circleci.com/gh/inpefess/residuated-binars.svg?style=svg\n   :target: https://circleci.com/gh/inpefess/residuated-binars\n.. |codecov| image:: https://codecov.io/gh/inpefess/residuated-binars/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/inpefess/residuated-binars\n.. |RTFD| image:: https://readthedocs.org/projects/residuated-binars/badge/?version=latest\n   :target: https://residuated-binars.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n.. |Binder| image:: https://mybinder.org/badge_logo.svg\n   :target: https://mybinder.org/v2/gh/inpefess/residuated-binars/HEAD?labpath=reproducing-residuated-binars-papers.ipynb\n.. |PyPI version| image:: https://badge.fury.io/py/residuated-binars.svg\n   :target: https://badge.fury.io/py/residuated-binars\n',
    'author': 'Boris Shminke',
    'author_email': 'boris@shminke.ml',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/inpefess/residuated-binars',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Algebraic Structure
====================
"""
from typing import Any, Dict, List, Union

CayleyTable = Dict[str, Dict[str, str]]
TOP = r"⟙"
BOT = r"⟘"


class AlgebraicStructure:
    r"""
        a base class for difference algebraic structures

    >>> magma_with_involution = AlgebraicStructure(
    ...     label="test",
    ...     operations={
    ...         "mult": {"0": {"0": "0", "1": "1"}, "1": {"0": "1", "1": "1"}},
    ...         "invo": {"0": "1", "1": "0"}
    ...     }
    ... )
    >>> magma_with_involution
    {'mult': [[0, 1], [1, 1]], 'invo': [1, 0]}
    >>> magma_with_involution.symbols
    ['0', '1']
    >>> magma_with_involution.remap_symbols({"0": "c1", "1": "c2"})
    >>> magma_with_involution
    {'mult': [[0, 1], [1, 1]], 'invo': [1, 0]}
    >>> magma_with_involution.symbols
    ['c1', 'c2']
    >>> print(magma_with_involution.mace4_format)
    mult(c1, c1) = c1.
    mult(c1, c2) = c2.
    mult(c2, c1) = c2.
    mult(c2, c2) = c2.
    invo(c1) = c2.
    invo(c2) = c1.
    <BLANKLINE>
    >>> class Magma(AlgebraicStructure):
    ...     ''' an example with redefined operation symbol '''
    ...
    ...     @property
    ...     def operation_map(self) -> Dict[str, str]:
    ...         return {"mult": "*"}
    >>> magma = Magma("test", {"mult": {"c": {"c": "c"}}})
    >>> print(magma.mace4_format)
    c * c = c.
    <BLANKLINE>
    """

    def __init__(
        self,
        label: str,
        operations: Dict[str, Dict[str, Any]],
    ):
        self.label = label
        self.operations = operations
        self.check_axioms()

    def check_axioms(self) -> None:
        """
        check axioms specific to that algebraic structure. If any axiom fails,
        raise an error

        :returns:
        """

    @property
    def cardinality(self) -> int:
        """
        :returns: number of items in the algebraic structure
        """
        return len(next(iter(self.operations.items()))[1].keys())

    def remap_symbols(self, symbol_map: Dict[str, str]) -> None:
        """rename symbols in a given way"""
        for op_label, operation in self.operations.items():
            if isinstance(next(iter(operation.items()))[1], Dict):
                new_table: CayleyTable = {}
                for one in symbol_map.keys():
                    new_table[symbol_map[one]] = {}
                    for two in symbol_map.keys():
                        new_table[symbol_map[one]][
                            symbol_map[two]
                        ] = symbol_map[operation[one][two]]
                self.operations[op_label] = new_table
            else:
                new_op: Dict[str, str] = {}
                for one in symbol_map.keys():
                    new_op[symbol_map[one]] = symbol_map[operation[one]]
                self.operations[op_label] = new_op

    @property
    def symbols(self) -> List[str]:
        """
        :returns: a list of symbols denoting items of an algebraic structure
        """
        keys = list(next(iter(self.operations.items()))[1].keys())
        pure_keys = keys.copy()
        if TOP in pure_keys:
            pure_keys.remove(TOP)
        if BOT in pure_keys:
            pure_keys.remove(BOT)
        return (
            ([BOT] if BOT in keys else [])
            + list(sorted(pure_keys))
            + ([TOP] if TOP in keys else [])
        )

    @property
    def operation_map(self) -> Dict[str, str]:
        """
        :returns: a map from operation labels (for functional notation) to
            operation symbols (for infix notation)
        """
        return {}

    @property
    def mace4_format(self) -> str:
        """
        represent the algebraic structure in ``Prover9/Mace4`` format
        :returns: a string representation
        """
        result = ""
        for op_label, operation in self.operations.items():
            op_symbol = self.operation_map.get(op_label, None)
            if isinstance(list(operation.items())[0][1], Dict):
                for i in self.symbols:
                    for j in self.symbols:
                        if op_symbol is not None:
                            result += (
                                f"{i} {op_symbol} {j} = {operation[i][j]}.\n"
                            )
                        else:
                            result += (
                                f"{op_label}({i}, {j}) = {operation[i][j]}.\n"
                            )
            else:
                for i in self.symbols:
                    result += f"{op_label}({i}) = {operation[i]}.\n"
        return result

    def _operation_tabular_view(
        self, operation: Dict[str, Any]
    ) -> Union[List[List[int]], List[int]]:
        inverse_index = {symbol: i for i, symbol in enumerate(self.symbols)}
        if isinstance(next(iter(operation.items()))[1], Dict):
            table: List[List[int]] = []
            for one in self.symbols:
                table.append([])
                for two in self.symbols:
                    table[inverse_index[one]].append(
                        inverse_index[operation[one][two]]
                    )
            return table
        new_operation: List[int] = list(range(len(operation)))
        for one in self.symbols:
            new_operation[inverse_index[one]] = inverse_index[operation[one]]
        return new_operation

    @property
    def tabular_format(self) -> Dict[str, Union[List[List[int]], List[int]]]:
        """
        :returns: a dictionary of Cayley tables as lists of lists
        """
        return {
            op_label: self._operation_tabular_view(operation)
            for op_label, operation in self.operations.items()
        }

    def __repr__(self):
        return str(self.tabular_format)



# Copyright 2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Pseudo-weak-:math:`R_0` Algebra
===============================
"""
from residuated_binars.algebraic_structure import BOT, TOP
from residuated_binars.axiom_checkers import (
    is_left_identity,
    left_distributive,
)
from residuated_binars.bounded_lattice import BoundedLattice


class PseudoWeakR0Algebra(BoundedLattice):
    """
    a representation of a peudo-weak-:math:`R_0` algebra

    for more info look `here <https://doi.org/10.1155/2014/854168>`__

    >>> join = {BOT: {BOT: BOT, TOP: TOP}, TOP: {BOT: TOP, TOP: TOP}}
    >>> meet = {BOT: {BOT: BOT, TOP: BOT}, TOP: {BOT: BOT, TOP: TOP}}
    >>> inv = {BOT: BOT, TOP: TOP}
    >>> imp = {BOT: {BOT: TOP, TOP: TOP}, TOP: {BOT: BOT, TOP: TOP}}
    >>> operations = {"imp1": imp, "imp2": imp, "inv1": inv, "inv2": inv,
    ...     "join": join, "meet": meet}
    >>> PseudoWeakR0Algebra("no P1", operations)
    Traceback (most recent call last):
     ...
    ValueError: P1 axiom doesn't hold
    >>> imp = {BOT: {BOT: TOP, TOP: TOP}, TOP: {BOT: TOP, TOP: TOP}}
    >>> operations = {"imp1": imp, "imp2": imp, "inv1": inv, "inv2": inv,
    ...     "join": join, "meet": meet}
    >>> PseudoWeakR0Algebra("no P1", operations)
    Traceback (most recent call last):
     ...
    ValueError: P2 axiom doesn't hold
    """

    def _check_p1(self) -> str:
        try:
            for one in self.operations["meet"].keys():
                for two in self.operations["meet"].keys():
                    assert (
                        self.operations["imp1"][one][two]
                        == self.operations["imp2"][
                            self.operations["inv1"][two]
                        ][self.operations["inv1"][one]]
                    )
                    assert (
                        self.operations["imp2"][one][two]
                        == self.operations["imp1"][
                            self.operations["inv2"][two]
                        ][self.operations["inv2"][one]]
                    )
            return " "
        except AssertionError:
            return "P1 axiom doesn't hold"

    def _check_p2(self) -> str:
        try:
            assert is_left_identity(self.operations["imp1"], TOP)
            assert is_left_identity(self.operations["imp2"], TOP)
            return " "
        except AssertionError:
            return "P2 axiom doesn't hold"

    def _check_p3(self) -> str:
        try:
            for i in self.operations["join"].keys():
                for j in self.operations["join"].keys():
                    for k in self.operations["join"].keys():
                        smaller = self.operations["imp1"][i][j]
                        greater = self.operations["imp1"][
                            self.operations["imp1"][k][i]
                        ][self.operations["imp1"][k][j]]
                        assert (
                            smaller == greater or smaller in self.more[greater]
                        )
                        smaller = self.operations["imp2"][i][j]
                        greater = self.operations["imp2"][
                            self.operations["imp2"][k][i]
                        ][self.operations["imp2"][k][j]]
                        assert (
                            smaller == greater or smaller in self.more[greater]
                        )
            return " "
        except AssertionError:
            return "P3 axiom doesn't hold"

    def _check_p4(self) -> str:
        try:
            assert left_distributive(
                self.operations["imp1"], self.operations["join"]
            )
            assert left_distributive(
                self.operations["imp2"], self.operations["join"]
            )
            return " "
        except AssertionError:
            return "P4 axiom doesn't hold"

    def check_axioms(self) -> None:
        super().check_axioms()
        assert (
            self.operations["inv1"][self.operations["inv2"][BOT]] == BOT
            and self.operations["inv2"][self.operations["inv1"][BOT]] == BOT
            and self.operations["inv1"][self.operations["inv2"][TOP]] == TOP
            and self.operations["inv2"][self.operations["inv1"][TOP]] == TOP
        ), "Pseudo-inverse axioms don't hold"
        res = " ".join(
            (
                self._check_p1(),
                self._check_p2(),
                self._check_p3(),
                self._check_p4(),
            )
        ).strip()
        if res != "":
            raise ValueError(res)
